import numpy as np
from milcapy.core.node import Node
from milcapy.section.section import ShellSection
from milcapy.utils.types import ConstitutiveModel

class MembraneQuad6:
    """Elemento de membrana cuadrilátero de 4 nodos + grados de libertad adicionales de perforacion (para compatibilidad de rigidez e evitar la inestabilidad numerica en el ensamblaje)
    Nota: Este elemento converge a la solucion excata de la viga de Euler-Bernoulli (sin considerar deformaciones por cortante)
    
    Args:
        id (int): Identificador del elemento.
        node1 (Node): Primer nodo.
        node2 (Node): Segundo nodo.
        node3 (Node): Tercer nodo.
        node4 (Node): Cuarto nodo.
        section (ShellSection): Sección del elemento tipo area (shell).
        state (ConstitutiveModel): Modelo constitutivo del elemento.
    
    Notas:
        Se recomienda usar este elemento por encima de los elementos MembraneQuad6I y MembraneQuad6IModificado.
    """
    def __init__(
        self,
        id: int,
        node1: Node,
        node2: Node,
        node3: Node,
        node4: Node,
        section: ShellSection,
        state: ConstitutiveModel,
    ) -> None:
        """
        Inicializa el elemento de membrana cuadrilátero de 4 nodos y 3 dof por nodo.
        """
        self.id = id
        self.node1 = node1
        self.node2 = node2
        self.node3 = node3
        self.node4 = node4
        self.section = section
        self.state = state
        self.E = self.section.E()
        self.v = self.section.v()
        self.t = self.section.t
        self.dofs = np.concatenate((node1.dofs, node2.dofs, node3.dofs, node4.dofs))
        r3 = np.sqrt(3)
        self.xi = [-1/r3, -1/r3, 1/r3, 1/r3]
        self.eta = [-1/r3, 1/r3, -1/r3, 1/r3]
        self.w = [1, 1, 1, 1]

    def set_gauss_points(self, xi: list[float], eta: list[float], w: list[float]):
        """
        Setea los puntos de gauss.
        """
        self.xi = xi
        self.eta = eta
        self.w = w

    def shape_function(self, xi: float, eta: float):
        """
        Funciones de forma.
        """
        N1 = (1-xi)*(1-eta)/4
        N2 = (1+xi)*(1-eta)/4
        N3 = (1+xi)*(1+eta)/4
        N4 = (1-xi)*(1+eta)/4
        N5 = (1-xi**2)*(1-eta)/2
        N6 = (1-eta**2)*(1+xi)/2
        N7 = (1-xi**2)*(1+eta)/2
        N8 = (1-eta**2)*(1-xi)/2
        return N1, N2, N3, N4, N5, N6, N7, N8

    def coordinates(self, xi: float, eta: float):
        """
        Interpolacion de coordenadas del elemento.
        """
        N1, N2, N3, N4, *_ = self.shape_function(xi, eta)
        x1, y1 = self.node1.coords
        x2, y2 = self.node2.coords
        x3, y3 = self.node3.coords
        x4, y4 = self.node4.coords
        x = N1*x1 + N2*x2 + N3*x3 + N4*x4
        y = N1*y1 + N2*y2 + N3*y3 + N4*y4
        return x, y

    def Ia(self) -> np.ndarray:
        """
        Matriz de unitaria ampliada.
        """
        Ia = np.array([
            [1, 0, 0, 0],
            [0, 0, 0, 1],
            [0, 1, 1, 0]
        ])
        return Ia

    def J(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz de jacobiano.
        """
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates
        J11 = -x1*(1-eta) + x2*(1-eta) + x3*(1+eta) - x4*(1+eta)
        J12 = -y1*(1-eta) + y2*(1-eta) + y3*(1+eta) - y4*(1+eta)
        J21 = -x1*(1-xi) - x2*(1+xi) + x3*(1+xi) + x4*(1-xi)
        J22 = -y1*(1-xi) - y2*(1+xi) + y3*(1+xi) + y4*(1-xi)
        return np.array([[J11, J12], [J21, J22]]) * (1/4)

    def Gamma(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz de jacobiano invertida.
        """
        J = self.J(xi, eta)
        gamma = np.linalg.inv(J)
        zero = np.zeros((2, 2))
        Gamma = np.concatenate((
            np.concatenate((gamma, zero), axis=1),
            np.concatenate((zero, gamma), axis=1)
        ),
            axis=0)
        return Gamma

    def dNdisp(self, xi: float, eta: float) -> np.ndarray:
        """
        Derivada de las funciones de forma (dof: desplazamientos).
        """
        dN = (1/4)*np.array([
            [eta-1, 0, 1-eta, 0, 1+eta, 0, -1-eta, 0],
            [xi-1,  0, -1-xi, 0, 1+xi,   0,  1-xi, 0],
            [0,     eta-1, 0, 1-eta, 0, 1+eta, 0, -1-eta],
            [0,     xi-1,  0, -1-xi, 0, 1+xi,   0,  1-xi]
        ])
        return dN

    def Bdisp(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz de deformacion de desplazamientos.
        """
        dN = self.dNdisp(xi, eta)
        Gamma = self.Gamma(xi, eta)
        Ia = self.Ia()
        B = Ia @ Gamma @ dN
        return B

    def dNrot(self, xi: float, eta: float) -> np.ndarray:
        """
        Derivada de las funciones de forma modificadas (dof: rotaciones).
        """
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates

        dN = (1/16)*np.array([
            [
                2 * xi * (eta - 1) * (y1 - y2)
                + (eta**2 - 1) * (y1 - y4),

                -2 * xi * (eta - 1) * (y1 - y2)
                - (eta**2 - 1) * (y2 - y3),

                -2 * xi * (eta + 1) * (y3 - y4)
                + (eta**2 - 1) * (y2 - y3),

                2 * xi * (eta + 1) * (y3 - y4)
                - (eta**2 - 1) * (y1 - y4)
            ],
            [
                2 * eta * (xi - 1) * (y1 - y4)
                + (xi**2 - 1) * (y1 - y2),

                -2 * eta * (xi + 1) * (y2 - y3)
                - (xi**2 - 1) * (y1 - y2),

                2 * eta * (xi + 1) * (y2 - y3)
                - (xi**2 - 1) * (y3 - y4),

                -2 * eta * (xi - 1) * (y1 - y4)
                + (xi**2 - 1) * (y3 - y4)
            ],
            [
                -2 * xi * (eta - 1) * (x1 - x2)
                - (eta**2 - 1) * (x1 - x4),

                2 * xi * (eta - 1) * (x1 - x2)
                + (eta**2 - 1) * (x2 - x3),

                2 * xi * (eta + 1) * (x3 - x4)
                - (eta**2 - 1) * (x2 - x3),

                -2 * xi * (eta + 1) * (x3 - x4)
                + (eta**2 - 1) * (x1 - x4)
            ],
            [
                -2 * eta * (x1 - x4) * (xi - 1)
                - (x1 - x2) * (xi**2 - 1),

                2 * eta * (x2 - x3) * (xi + 1)
                + (x1 - x2) * (xi**2 - 1),

                -2 * eta * (x2 - x3) * (xi + 1)
                + (x3 - x4) * (xi**2 - 1),

                2 * eta * (x1 - x4) * (xi - 1)
                - (x3 - x4) * (xi**2 - 1)
            ]
        ])
        return dN

    def Brot(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz de deformacion de rotaciones.
        """
        dN = self.dNrot(xi, eta)
        Gamma = self.Gamma(xi, eta)
        Ia = self.Ia()
        B = Ia @ Gamma @ dN
        return B

    def B(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz de deformacion.
        """
        B = np.hstack((self.Bdisp(xi, eta), self.Brot(xi, eta)))

        return B

    def D(self) -> np.ndarray:
        """
        Matriz constitutiva de esfuerzo plano.
        """
        v = self.section.v()
        E = self.section.E()
        # ! T E N S I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRESS:
            k = E/(1-v**2)
            return k*np.array([
                            [1, v, 0],
                            [v, 1, 0],
                            [0, 0, (1-v)/2],
                            ])

        # ! D E F O R M A C I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRAIN:
            k = (E*(1-v))/((1 + v)*(1 - 2*v))
            return k * np.array([
                [1,                 v/(1-v), 0],
                [v/(1-v),           1,       0],
                [0,                 0,       (1-2*v)/(2*(1-v))],
            ])

    def phiKi(self, xi: float, eta: float):
        """
        Matriz funcional de rigidez.
        """
        detJ = np.linalg.det(self.J(xi, eta))
        B = self.B(xi, eta)
        D = self.D()
        t = self.t
        return B.T @ D @ B * detJ * t

    def Ki(self):
        """
        Calcula la matriz rigidez atravez de la intergracion numerica cuadratura de gauss.
        """
        Ki = np.zeros((12, 12))
        for i in range(len(self.xi)):
            Ki += self.phiKi(self.xi[i], self.eta[i]) * self.w[i]


        # orden actual: [u1x, u1y, u2x, u2y, u3x, u3y, u4x, u4y, r1z, r2z, r3z, r4z]
        # orden deseado: [u1x, u1y, r1z, u2x, u2y, r2z, u3x, u3y, r3z, u4x, u4y, r4z]

        perm = [0, 1, 8, 2, 3, 9, 4, 5, 10, 6, 7, 11]

        # reordenar filas y columnas
        Ki = Ki[np.ix_(perm, perm)]


        return Ki #[U1X, U1Y, r1Z,    U2X, U2Y, r2Z,    U3X, U3Y, r3Z,    U4X, U4Y, r4Z]

    def get_transformation_matrix(self):
        """
        Matriz de transformacion 12x12 de local a global.
        """
        angle = np.arctan2(self.node2.vertex.y - self.node1.vertex.y, self.node2.vertex.x - self.node1.vertex.x)

        T = np.array([
            [np.cos(angle), -np.sin(angle), 0,      0,              0,        0,      0,              0,        0,      0,              0,        0],
            [np.sin(angle),  np.cos(angle), 0,      0,              0,        0,      0,              0,        0,      0,              0,        0],
            [0,                   0,        1,      0,              0,        0,      0,              0,        0,      0,              0,        0],
            [0,                   0,        0, np.cos(angle), -np.sin(angle), 0,      0,              0,        0,      0,              0,        0],
            [0,                   0,        0, np.sin(angle),  np.cos(angle), 0,      0,              0,        0,      0,              0,        0],
            [0,                   0,        0,      0,              0,        1,      0,              0,        0,      0,              0,        0],
            [0,                   0,        0,      0,              0,        0, np.cos(angle), -np.sin(angle), 0,      0,              0,        0],
            [0,                   0,        0,      0,              0,        0, np.sin(angle),  np.cos(angle), 0,      0,              0,        0],
            [0,                   0,        0,      0,              0,        0,      0,              0,        1,      0,              0,        0],
            [0,                   0,        0,      0,              0,        0,      0,              0,        0, np.cos(angle), -np.sin(angle), 0],
            [0,                   0,        0,      0,              0,        0,      0,              0,        0, np.sin(angle),  np.cos(angle), 0],
            [0,                   0,        0,      0,              0,        0,      0,              0,        0,      0,              0,        1],

        ])
        return T


    def global_stiffness_matrix(self):
        """
        Calcula la matriz rigidez global.
        """
        # T = self.get_transformation_matrix()
        Ki = self.Ki()
        # K_GLOBAL = T @ Ki @ T.T
        K_GLOBAL = Ki
        return K_GLOBAL




if __name__ == "__main__":
    import pandas as pd
    pd.set_option('display.float_format', '{:.5f}'.format)
    E = 25  # MPa
    v = 0.20
    t = 400  # mm
    l = 2000  # mm
    h = 600  # mm
    state = ConstitutiveModel.PLANE_STRESS
    from milcapy.utils.geometry import Vertex
    from milcapy.section.section import ShellSection
    from milcapy.material.material import Material

    material = Material("mat",E, v, 0)
    section = ShellSection("sec",material, t)
    nd1 = Node(1, Vertex(0, 0))
    nd2 = Node(2, Vertex(l, 0))
    nd3 = Node(3, Vertex(l, h))
    nd4 = Node(4, Vertex(0, h))
    el = MembraneQuad6(1, nd1, nd2, nd3, nd4, section, state)
    k = el.global_stiffness_matrix()
    f = 20
    # q = np.array([0, -0.5*f, 0,-0.5*f, 0, 0, 0, 0])
    q = np.array([0,     0, -0.5*f, 0,    0, -0.5*f, 0,    0])
    idx =        [3,     4, 5, 6,         7, 8, 9,         12]
    idx = np.array(idx) - 1
    k =k[np.ix_(idx, idx)]

    u = np.linalg.solve(k, q)
    # print(pd.DataFrame(k))
    print(pd.DataFrame(u))


from .types import (
    Restraints
)

from .types import (
    MemberType,
    CoordinateSystemType,
    DirectionType,
    StateType,
    LoadType,
    BeamTheoriesType,
    InternalForceType,
    to_enum,
    IntegrationType
)

from ..utils.geometry import (
    MatrixException,
    rotation_matrix,
    rotate_xy,
    traslate_xy,
    angle_x_axis,
    vertex_range
)




from abc import ABC, abstractmethod
from milcapy.material.material import Material
from math import pi
from milcapy.utils.types import ShearCoefficientMethodType

class Section(ABC):
    """
    Clase base para representar una sección estructural.

    Esta clase proporciona los métodos y propiedades comunes para todas las secciones
    estructurales. Las secciones estructurales son objetos que representan una sección
    transversal de un miembro estructural, como una viga o una columna. Las secciones
    estructurales tienen un nombre, un material, un método de cálculo del coeficiente de
    corte y propiedades geométricas como el área de la sección transversal y el momento
    de inercia.

    Attributes:
        name (str): Nombre de la sección.
        material (Material): Material asociado a la sección.
        shear_method (ShearCoefficientMethodType): Método de cálculo del coeficiente de corte.
        kA (float): Modificador de área. Por defecto es 1.
        kAs (float): Modificador de área para cálculo de la rigidez a cortante. Por defecto es 1.
        kI (float): Modificador de momento de inercia. Por defecto es 1.
        kg (float): Modificador de peso. Por defecto es 1.

    Properties:
        kg (float): Modificador de peso.
        v (float): Coeficiente de Poisson.
        E (float): Módulo de elasticidad.
        g (float): Peso específico.
        G (float): Módulo de corte.

    Abstract methods:
        A (float): Área de la sección transversal.
        I (float): Momento de inercia de la sección.
        k (float): Coeficiente de corte o Timoshenko de la sección.

    """
    def __init__(
        self,
        name: str,
        material: Material,
        shear_method: ShearCoefficientMethodType
    ):
        """
        Inicializa una sección estructural.

        Args:
            name (str): Nombre de la sección.
            material (Material): Material asociado a la sección.
            shear_method (ShearCoefficientMethodType): Método de cálculo del coeficiente de corte.
        """
        self.name = name
        self.material = material
        self.shear_method = shear_method
        self.kA = 1
        self.kAs = 1
        self.kI = 1
        self.kg = 1

    @property
    def kg(self) -> float:
        """Modificador de peso."""
        return self.material.kg

    @kg.setter
    def kg(self, value: float) -> None:
        self.material.kg = value

    def v(self) -> float:
        """Coeficiente de Poisson."""
        return self.material.v

    def E(self) -> float:
        """Módulo de elasticidad."""
        return self.material.E

    def g(self) -> float:
        """Peso específico."""
        return self.material.g * self.kg

    def G(self) -> float:
        """Módulo de corte."""
        return self.material.G

    @abstractmethod
    def A(self) -> float:
        """Área de la sección transversal."""
        pass

    @abstractmethod
    def I(self) -> float:
        """Momento de inercia de la sección."""
        pass

    @abstractmethod
    def k(self) -> float:
        """Coeficiente de corte o Timoshenko de la sección."""
        pass

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: name={self.name}, material=({self.material.name}), A={self.A():.3f}, I={self.I():.3f}, k={self.k():.3f}, kg={self.kg():.3f}"


class RectangularSection(Section):
    """
    Clase para representar una sección rectangular.
    """
    def __init__(
        self,
        name: str,
        material: Material,
        base: float,
        height: float,
        shear_method: ShearCoefficientMethodType
    ):
        """
        Inicializa una sección rectangular.

        Args:
            name (str): Nombre de la sección.
            material (Material): Material asociado a la sección.
            base (float): Base de la sección (en unidades de longitud).
            height (float): Altura de la sección (en unidades de longitud).
            shear_method (ShearCoefficientMethodType): Método de cálculo del coeficiente de corte.

        Raises:
            ValueError: Si la base o la altura son menores o iguales a cero.
        """
        if base <= 0 or height <= 0:
            raise ValueError("La base y la altura deben ser mayores que cero.")
        super().__init__(name, material, shear_method)
        self.base = base
        self.height = height
        self.shear_method = shear_method

    def A(self) -> float:
        """Área de la sección transversal."""
        return self.base * self.height * self.kA

    def I(self) -> float:
        """Momento de inercia de la sección."""
        return (self.base * self.height ** 3) / 12 * self.kI

    def k(self) -> float:
        """
        Coeficiente de corte de Timoshenko de la sección.

        Returns:
            float: Coeficiente de corte

        Raises:
            ValueError: Si el método no es válido
        """
        if self.shear_method == ShearCoefficientMethodType.TIMOSHENKO:
            return 5/6 * self.kAs
        elif self.shear_method == ShearCoefficientMethodType.COWPER:
            return 10 * (1 + self.v()) / (12 + 11 * self.v()) * self.kAs
        else:
            raise ValueError(f"Método de coeficiente no válido: {self.shear_method}. Opciones válidas: 'TIMOSHENKO', 'COWPER'")


class CircularSection(Section):
    """
    Clase para secciones circulares.
    """
    def __init__(
        self,
        name: str,
        material: Material,
        radius: float,
        shear_method: ShearCoefficientMethodType
    ):
        """
        Inicializa una sección circular.

        Args:
            name (str): Nombre de la sección.
            material (Material): Material asociado a la sección.
            radius (float): Radio de la sección.
            shear_method (ShearCoefficientMethodType): Método de cálculo del coeficiente de corte.
        Raises:
            ValueError: Si el radio es menor o igual a cero.
        """
        if radius <= 0:
            raise ValueError("El radio debe ser positivo.")
        super().__init__(name, material, shear_method)
        self.radius = radius
        self.shear_method = shear_method

    def A(self) -> float:
        """Área de la sección transversal."""
        return pi * self.radius ** 2 * self.kA

    def I(self) -> float:
        """Momento de inercia de la sección."""
        return (pi * self.radius ** 4) / 4 * self.kI

    def k(self) -> float:
        """
        Coeficiente de corte o Timoshenko de la sección.

        Returns:
            float: Coeficiente de corte

        Raises:
            ValueError: Si el método no es válido
        """
        if self.shear_method == ShearCoefficientMethodType.TIMOSHENKO:
            return 6/7 * self.kAs
        elif self.shear_method == ShearCoefficientMethodType.COWPER:
            return 10 * (1 + self.v()) / (12 + 11 * self.v()) * self.kAs
        else:
            raise ValueError(f"Método de coeficiente no válido: {self.shear_method}. Opciones válidas: 'TIMOSHENKO', 'COWPER'")


class GenericSection(Section):
    """Clase para secciones genéricas con propiedades básicas."""

    def __init__(
        self,
        name: str,
        material: Material,
        area: float,
        inertia: float,
        k_factor: float,
    ):
        """
        Inicializa una sección genérica.

        Args:
            name (str): Nombre de la sección.
            material (Material): Material asociado a la sección.
            area (float): Área de la sección.
            inertia (float): Momento de inercia de la sección.
            k_factor (float): Coeficiente de corte de la sección.

        Raises:
            ValueError: Si el área, el momento de inercia o el coeficiente de corte son inválidos.
        """
        if area <= 0 or inertia <= 0 or k_factor <= 0:
            raise ValueError("El área, el momento de inercia y el coeficiente de corte deben ser mayores que cero.")
        super().__init__(name, material, ShearCoefficientMethodType.NONE)
        self._area = area
        self._inertia = inertia
        self._k_factor = k_factor

    def A(self) -> float:
        """Área de la sección transversal."""
        return self._area * self.kA

    def I(self) -> float:
        """Momento de inercia de la sección."""
        return self._inertia * self.kI

    def k(self) -> float:
        """
        Coeficiente de corte de Timoshenko de la sección.

        Returns:
            float: Coeficiente de corte
        """
        return self._k_factor * self.kAs


class ShellSection:
    """Clase para secciones de cascaras"""
    def __init__(
        self,
        name: str,
        material: Material,
        thickness: float,
    ):
        """
        Inicializa una sección de cascaras.

        Args:
            name (str): Nombre de la sección.
            material (Material): Material asociado a la sección.
            thickness (float): Grosor de la sección.

        Raises:
            ValueError: Si el grosor es inválido.
        """
        if thickness <= 0:
            raise ValueError("El grosor debe ser mayor que cero.")
        self.name = name
        self.material = material
        self.t = thickness

    def E(self):
        return self.material.E
    def v(self):
        return self.material.v
    def G(self):
        return self.material.G
    def g(self):
        return self.material.g
    def alpha(self):
        return self.material.alpha
    def beta(self):
        return self.material.beta
    def gamma(self):
        return self.material.gamma


from milcapy.elements.CST import MembraneTriangle
from typing import Optional
import numpy as np

class PP_CST(MembraneTriangle):
    """
    Un postprocesador para el elemento "Constant Straint Triangle" (CST)

    Permite obtener las deformaciones y tensiones en cada nodo del elemento CST

    Attributes:
        displacements (Optional[np.ndarray]): array con los valores de desplazamiento en cada nodo [ux1, uy1, ux2, uy2, ux3, uy3]
        coordinates (np.ndarray): array con las coordenadas de cada nodo [x1, y1, x2, y2, x3, y3]
    """
    def __init__(self):
        self.displacements: Optional[np.ndarray] = None

    def strains(self):
        """
        Calcula las deformaciones en cada nodo del elemento CST

        Returns:
            np.ndarray: array con las deformaciones en cada nodo
        """
        return self.B @ self.displacements

    def stresses(self):
        """
        Calcula las tensiones en cada nodo del elemento CST

        Returns:
            np.ndarray: array con las tensiones en cada nodo
        """
        return self.D @ self.strains()

    def process_builder(self, cst: MembraneTriangle, displacements: np.ndarray):
        """
        Crea un nuevo objeto PP_CST y asigna los valores de desplazamiento y coordenadas

        Args:
            cst (MembraneTriangle): objeto CST del que se va a obtener la informacion
            displacements (np.ndarray): array con los valores de desplazamiento en cada nodo
        """
        super().__init__(0, cst.node1, cst.node2, cst.node3, cst.section, cst.state)
        self.displacements = displacements
        self.coordinates = np.concatenate(([cst.node1.vertex.coordinates, cst.node2.vertex.coordinates, cst.node3.vertex.coordinates]))

    def deformed_shape(self) -> np.ndarray:
        """
        Calcula la forma deformada del elemento CST

        Returns:
            np.ndarray: array con las coordenadas de la forma deformada
        """
        return self.coordinates + self.displacements


from milcapy.elements.member import Member
from milcapy.elements.truss import TrussElement
from typing import Dict, Tuple, Optional, Union
import numpy as np
from milcapy.utils.element import q_phi, length_offset_transformation_matrix

class BeamSeg():
    """
    Un segmento de viga matemáticamente continuo
    """

    def __init__(self):
        self.xi: float | None = None  # Ubicación inicial del segmento de la viga (relativa al inicio de la viga)
        self.xj: float | None = None  # Ubicación final del segmento de la viga (relativa al inicio de la viga)
        self.qi: float | None = None  # Carga distribuida transversal lineal en el inicio del segmento
        self.qj: float | None = None  # Carga distribuida transversal lineal en el final del segmento
        self.pi: float | None = None  # Carga distribuida axial lineal en el inicio del segmento
        self.pj: float | None = None  # Carga distribuida axial lineal en el final del segmento
        self.Pi: float | None = None  # Fuerza axial interna en el inicio del segmento
        self.Pj: float | None = None  # Fuerza axial interna en el final del segmento
        self.Vi: float | None = None  # Fuerza cortante interna en el inicio del segmento
        self.Vj: float | None = None  # Fuerza cortante interna en el final del segmento
        self.Mi: float | None = None  # Momento interno en el inicio del segmento
        self.Mj: float | None = None  # Momento interno en el final del segmento
        self.ui: float | None = None  # Desplazamiento axial en el inicio del segmento de la viga
        self.uj: float | None = None  # Desplazamiento axial en el final del segmento de la viga
        self.vi: float | None = None  # Desplazamiento transversal en el inicio del segmento de la viga
        self.vj: float | None = None  # Desplazamiento transversal en el final del segmento de la viga
        self.thetai: float | None = None  # Pendiente en el inicio del segmento de la viga
        self.thetaj: float | None = None  # Pendiente en el final del segmento de la viga
        self.E: float | None = None  # Módulo de elasticidad
        self.I: float | None = None  # Inercia
        self.A: float | None = None  # Área
        self.phi: float | None = None  # Aporte por cortante
        # Coeficientes de integración
        self.C1: float | None = None
        self.C2: float | None = None
        self.C3: float | None = None
        self.C4: float | None = None

        # TOPICOS
        # Brazos rigidos
        self.la: float | None = None
        self.lb: float | None = None
        self.qla: bool | None = None
        self.qlb: bool | None = None
        self.member: Optional[Union[Member, TrussElement]] | None = None

    def coefficients(self):
        """
        Retorna los coeficientes de integración
        """
        L = self.le()
        qi = self.qi
        qj = self.qj
        E = self.E
        I = self.I
        phi = self.phi
        vi = self.vi
        vj = self.vj
        thetai = self.thetai
        thetaj = self.thetaj

        A = -(qj - qi) / L
        B = -qi

        M = np.array([
            [0,                     0,          0, 1],
            [0,                     0,          1, 0],
            [L**3 * (2 - phi) / 12, L**2 / 2,   L, 1],
            [L**2 / 2,              L,          1, 0],
        ])

        N = np.array([
            E * I * vi,
            E * I * thetai,
            E * I * vj + A * L**5 * (0.6 - phi) / 72 + B * L**4 * (1 - phi) / 24,
            E * I * thetaj + A * L**4 / 24 + B * L**3 / 6,
        ])

        C = np.linalg.solve(M, N)

        self.C1 = C[0]
        self.C2 = C[1]
        self.C3 = C[2]
        self.C4 = C[3]

        return tuple(map(float, C))

    def length(self):
        """
        retorna la longitud total del elemento
        """
        return self.xj - self.xi

    def le(self):
        """
        Retorna la longitud del segmento flexible
        """

        return self.xj - self.xi - (self.la or 0) - (self.lb or 0)


    def shear(self, x):
        """
        Retorna la fuerza cortante en un punto 'x' del segmento
        """

        la, lb, le = self.la or 0, self.lb or 0, self.le()
        if (self.la or self.lb) and (0 <= x < la or le+la < x <= la+le+lb):
            return 0

        Vi = self.Vi
        qi = self.qi
        qj = self.qj
        L = self.le()
        A = (qj - qi)/L
        B = qi
        x = x - la

        return Vi + B*x + A*x**2/2

    def moment(self, x):
        """
        Retorna el momento en un punto 'x' del segmento
        """

        la, lb, le = self.la or 0, self.lb or 0, self.le()
        if (self.la or self.lb) and (0 <= x < la or le+la < x <= la+le+lb):
            return 0

        Vi = self.Vi
        Mi = self.Mi
        qi = self.qi
        qj = self.qj
        L = self.le()
        A = (qj - qi)/L
        B = qi
        x = x - la

        M = Mi - Vi*x - B*x**2/2 - A*x**3/6

        return M

    def axial(self, x):
        """
        Retorna la fuerza axial en un punto 'x' del segmento
        """

        la, lb, le = self.la or 0, self.lb or 0, self.le()
        if (self.la or self.lb) and (0 <= x < la or le+la < x <= la+le+lb):
            return 0

        C1 = - self.Pi
        A = (self.pj - self.pi)/le
        B = self.pi
        x = x - la

        N = - A*x**2/2 - B*x + C1

        return N

    def axial_displacement(self, x):
        """
        Retorna el desplazamiento axial en un punto 'x' del segmento
        """

        la, lb, le = self.la or 0, self.lb or 0, self.le()
        if (self.la or self.lb) and (0 <= x < la):
            return self.ui
        if (self.la or self.lb) and (le+la < x <= la+le+lb):
            return self.uj

        L = self.le()
        EA = self.E * self.A
        EAoverL = EA / L
        C2 = self.ui * EA
        A = (self.pi - self.pj)/L
        B = self.pi
        C1 = EAoverL * (self.uj - self.ui) + L * (B - A * L/3)
        x = x - la

        u = 1/EA * (- A*x**3/6 - B*x**2/2 + C1*x + C2)

        return u

    def slope(self, x):
        """
        Retorna la pendiente de la curva elástica en cualquier punto `x` a lo largo del segmento.
        """

        la, lb, le = self.la or 0, self.lb or 0, self.le()
        if (self.la or self.lb) and (0 <= x < la):
            return self.thetai
        if (self.la or self.lb) and (le+la < x <= la+le+lb):
            return self.thetaj

        qi = self.qi
        qj = self.qj
        L = self.le()
        EI = self.E * self.I
        A = (qj - qi)/L
        B = qi

        C1 = self.C1
        C2 = self.C2
        C3 = self.C3
        x = x - la

        theta_x = 1/EI * (A*x**4/24 + B*x**3/6 + C1*x**2/2 + C2*x + C3)

        return theta_x

    def deflection(self, x):
        """
        Retorna la deflexión en un punto 'x' del segmento
        """

        la, lb, le = self.la or 0, self.lb or 0, self.le()
        if (self.la or self.lb):
            if (0 <= x < la):
                vi = self.vi - la * self.thetai
                return vi + x * self.thetai
            elif (le+la < x <= la+le+lb):
                return self.vj + (x - le - la) * self.thetaj

        qi = self.qi
        qj = self.qj
        L = self.le()
        EI = self.E * self.I
        A = (qj - qi)/L
        B = qi
        x = x - la
        phi = self.phi

        C1 = self.C1
        C2 = self.C2
        C3 = self.C3
        C4 = self.C4
        term1 = A * L**2 * x**3 * (0.6 * (x / L)**2 - phi) / 72
        term2 = B * L**2 * x**2 * ((x / L)**2 - phi) / 24
        term3 = C1 * x * L**2 * (2 * (x / L)**2 - phi) / 12
        term4 = C2 * x**2 / 2
        term5 = C3 * x
        term6 = C4

        # C1 = self.Vi
        # C2 = self.Mi
        # C3 = self.thetai
        # C4 = self.vi
        # term1 = - A * x**5 / 120
        # term2 = - B * x**4 / 24
        # term3 = (12 * C1 + phi * A * L**2) * x**3 / 72
        # term4 = (12 * C2 + phi * B * L**2) * x**2 / 24
        # term5 = C3 * x
        # term6 = C4

        return (term1 + term2 + term3 + term4 + term5 + term6) / (EI)


    def process_builder(self, member: "Member", results: "Dict[str, np.ndarray]", pattern_name: str) -> None:

        """
        Inicializa las variables (Propiedades para el calculo de respuestas a lo largo del elemento marco)
        """

        self.la = member.la
        self.lb = member.lb
        self.qla = member.qla
        self.qlb = member.qlb
        self.fla = member.fla
        self.flb = member.flb

        self.xi =0
        self.xj = member.length()

        if self.la or self.lb:
            # NOTE: si tiene fla, flb este se calcula en los extremos efectivos del elementos flexible (L = Length -la*fla -lb*flb)
            # NOTE: ya luego con condicionales de toma los valores desde la y lb (dentro de ellos pone 0)
            local_disp = results["displacements"]
            # LO_T = member.length_offset_transformation_matrix()
            LO_T = length_offset_transformation_matrix(member.la or 0, member.lb or 0)
            local_displacements_flex = LO_T @ local_disp

            load = member.get_distributed_load(pattern_name)
            L = member.length()
            la, lb = self.la or 0, self.lb or 0
            # ###### CORRECCION POR FLA Y FLB: en el proceso se hace con la*fla y lb*flb
            # # mas en postprocess se hace con la y lb
            # if self.fla or self.flb:
            #     la, lb = member.la_lb_efect()
            # ####################
            le = L - la - lb
            qi, qj, pi, pj = load.q_i, load.q_j, load.p_i, load.p_j


            a = (qj - qi) / L
            b = qi

            c = (pj - pi) / L
            d = pi

            qa = a*la + b
            qb = a*(L -lb) + b

            pa = c*la + d
            pb = c*(L -lb) + d

            ####### CORRECIION ADICIONAL: el usuario al darle qla, qlb == False
            # indica que los valores qi, qj, pi, pj ingresados son en la cara del brazo rigido
            # por lo que se debria de crar un nuevo omdelo de carga PartialLoad pero en esta ocacion solo modificaremos el vector de cargas
            if self.qla == False:
                qa = qi
                pa = pi
            if self.qlb == False:
                qb = qj
                pb = pj
            ####################



            load_vector = q_phi(le, member.phi(), qa, qb, pa, pb)
            stiffness_matrix = member.flexible_stiffness_matrix()
            internal_forces_flex = stiffness_matrix @ local_displacements_flex - load_vector
            fixed_end_forces = [qa, qb, pa, pb]

        else:
            load = member.get_distributed_load(pattern_name)
            qi, qj, pi, pj = load.q_i, load.q_j, load.p_i, load.p_j

            local_displacements_flex = results["displacements"]
            internal_forces_flex = results["internal_forces"]
            fixed_end_forces = [qi, qj, pi, pj]

        # TODAS LAS ACCIONES Y REACCIONES EN LOS EXTREMOS DE LA PARTE FLEXIBLE
        self.qi = fixed_end_forces[0]
        self.qj = fixed_end_forces[1]
        self.pi = fixed_end_forces[2]
        self.pj = fixed_end_forces[3]
        self.Pi = internal_forces_flex[0]
        self.Pj = internal_forces_flex[3]
        self.Vi = internal_forces_flex[1]
        self.Vj = internal_forces_flex[4]
        self.Mi = internal_forces_flex[2]
        self.Mj = internal_forces_flex[5]
        self.ui = local_displacements_flex[0]
        self.uj = local_displacements_flex[3]
        self.vi = local_displacements_flex[1]
        self.vj = local_displacements_flex[4]
        self.thetai = local_displacements_flex[2]
        self.thetaj = local_displacements_flex[5]
        self.E = member.section.E()
        self.I = member.section.I()
        self.A = member.section.A()
        self.phi = member.phi()
        self.member = member


    def process_builder_for_truss(self, truss: "TrussElement", results: dict, pattern_name: str):
        self.xi = 0
        self.xj = truss.length()
        self.qi = 0
        self.qj = 0
        self.pi = truss.get_distributed_load(pattern_name).p_i
        self.pj = truss.get_distributed_load(pattern_name).p_j
        self.Pi = results["internal_forces"][0]
        self.Pj = results["internal_forces"][1]
        self.Vi = 0
        self.Vj = 0
        self.Mi = 0
        self.Mj = 0
        self.ui = results["displacements"][0]
        self.uj = results["displacements"][1]
        self.vi = 0
        self.vj = 0
        self.thetai = 0
        self.thetaj = 0
        self.E = truss.section.E()
        self.I = 0
        self.A = truss.section.A()
        self.phi = 0
        # Coeficientes de integración
        self.C1 = 0
        self.C2 = 0
        self.C3 = 0
        self.C4 = 0

        # TOPICOS
        # Brazos rigidos
        self.la = 0
        self.lb = 0
        self.qla = True
        self.qlb = True

def deformed_shape(member: "Member", results: dict, escale: float) -> Tuple[np.ndarray, np.ndarray]:

    """Calcula las coordenada de la deformada en sistema local"""

    L = member.length()

    flecha = results["deflections"] * escale

    x = results["x_val"]
    x_val = results["axial_displacements"]*escale + x

    return x_val, flecha


from typing import TYPE_CHECKING, Dict, Optional
from milcapy.loads.load import PointLoad, PrescribedDOF, ElasticSupport, LocalAxis
import numpy as np

if TYPE_CHECKING:
    from milcapy.utils.types import Restraints
    from milcapy.utils.geometry import Vertex

class NodeArbitrary:
    def __init__(
        self,
        id: int,
        vertex: "Vertex"
    ) -> None:
        self.id: int = id
        self.vertex: "Vertex" = vertex

class Node:
    """
    Representa un nodo en el modelo estructural.
    """

    def __init__(
        self,
        id: int,
        vertex: "Vertex"
    ) -> None:
        """
        Inicializa un nodo con su identificador y posición.

        Args:
            id (int): Identificador único del nodo.
            vertex (Vertex): Coordenadas (x, y) del nodo.
        """
        self.id: int = id
        self.vertex: "Vertex" = vertex
        self.restraints: "Restraints" = (
            False, False, False)  # Restricciones del nodo

        # Cálculo de los índices de grados de libertad
        self.dofs: np.ndarray = np.array([
            self.id * 3 - 2,  # DOF en x
            self.id * 3 - 1,  # DOF en y
            self.id * 3       # DOF en theta
        ], dtype=int)

        # Cargas aplicadas al nodo
        self.loads: Dict[str, PointLoad] = {}   # {pattern_name: PointLoad}

        # Patrón de carga actual
        self.current_load_pattern: Optional[str] = None

        # Desplazamientos prescindidos
        self.prescribed_dofs: Dict[str, PrescribedDOF] = {}

        # apoyos elásticos
        self.elastic_supports: Optional[ElasticSupport] = None

        # eje local
        self.local_axis: Optional[LocalAxis] = None

    def load_vector(self) -> np.ndarray:
        """
        Vector de cargas aplicadas al nodo para el patrón de carga actual en sistema global.
        """
        load = self.loads.get(self.current_load_pattern, PointLoad())
        return load.components

    def set_current_load_pattern(self, load_pattern_name: str) -> None:
        """
        Establece el patrón de carga actual del nodo.
        """
        self.current_load_pattern = load_pattern_name

    def set_restraints(self, restraints: "Restraints") -> None:
        """
        Establece las restricciones del nodo.
        """
        self.restraints = restraints

    def set_load(self, load: "PointLoad") -> None:
        """
        Establece una carga puntual en el nodo.
        """
        self.loads[self.current_load_pattern] = load

    def set_prescribed_dof(self, prescribed_dof: "PrescribedDOF") -> None:
        """
        Establece los desplazamientos prescindidos del nodo.
        """
        self.prescribed_dofs[self.current_load_pattern] = prescribed_dof

    def set_elastic_support(self, elastic_support: ElasticSupport) -> None:
        """
        Establece los apoyos elásticos del nodo.
        """
        self.elastic_supports = elastic_support

    def set_local_axis(self, local_axis: LocalAxis) -> None:
        """
        Establece el eje local del nodo. ojo esto hace que el vector de fuerzas aplicados, desplazamientos prescindidos y apoyos elásticos se transformen al sistema local.
        """
        self.local_axis = local_axis



"""
==============================================
ELEMENTO CUADRILATERO MEMBRANA (Q8)
==============================================
|| Nodos                 || 8               ||
|| DOF por nodo          || 2               ||
|| Integracion           || Reducida 4P     ||
|| Shape Function        || Bi Cuadrada     ||
|| Formulacion           || Isoparametrica  ||
|| clase de elemento     || Serendipitos    ||
|| Estado de programacion|| Listo           ||
==============================================
"""


from milcapy.core.node import Node, NodeArbitrary
from milcapy.section.section import ShellSection
import numpy as np
import uuid
from milcapy.utils.types import ConstitutiveModel, IntegrationType
from milcapy.utils.geometry import Vertex

class MembraneQuad8:
    def __init__(
        self,
        id: int,
        node1: Node,
        node2: Node,
        node3: Node,
        node4: Node,
        section: ShellSection,
        state: ConstitutiveModel,
        integration: IntegrationType
    ) -> None:
        self.id = id
        self.node1 = node1
        self.node2 = node2
        self.node3 = node3
        self.node4 = node4
        self.node5 = NodeArbitrary(uuid.uuid4().int, Vertex(*((node1.vertex.coordinates + node2.vertex.coordinates)/2)))
        self.node6 = NodeArbitrary(uuid.uuid4().int, Vertex(*((node2.vertex.coordinates + node3.vertex.coordinates)/2)))
        self.node7 = NodeArbitrary(uuid.uuid4().int, Vertex(*((node3.vertex.coordinates + node4.vertex.coordinates)/2)))
        self.node8 = NodeArbitrary(uuid.uuid4().int, Vertex(*((node4.vertex.coordinates + node1.vertex.coordinates)/2)))
        self.state = state
        self.integration = integration
        self.section = section
        self.E = section.E()
        self.v = section.v()
        self.t = section.t
        self.dofs = np.concatenate(
            (node1.dofs[:2], node2.dofs[:2], node3.dofs[:2], node4.dofs[:2]))

        if self.integration == IntegrationType.REDUCED:
            r3 = np.sqrt(3)
            self.xi = [-1/r3, -1/r3, 1/r3, 1/r3]
            self.eta = [-1/r3, 1/r3, -1/r3, 1/r3]
            self.w = [1, 1, 1, 1]
        elif self.integration == IntegrationType.COMPLETE:
            ra = np.sqrt(0.6)
            self.xi = [-ra, ra, ra, -ra, 0, ra, 0, -ra, 0]
            self.eta = [-ra, -ra, ra, ra, -ra, 0, ra, 0, 0]
            self.w = [25/81, 25/81, 25/81, 25/81, 40/81, 40/81, 40/81, 40/81, 64/81]

    def shape_function(self, xi: float, eta: float) -> tuple[float, float, float, float, float, float, float, float]:
        """
        Calcula las funciones de forma en un punto (xi, eta).
        """
        N1 = -1/4 * (1-xi) * (1-eta) * (1+xi+eta)
        N2 = -1/4 * (1+xi) * (1-eta) * (1-xi+eta)
        N3 = -1/4 * (1+xi) * (1+eta) * (1-xi-eta)
        N4 = -1/4 * (1-xi) * (1+eta) * (1+xi-eta)
        N5 = 1/2 * (1-xi) * (1-eta) * (1+xi)
        N6 = 1/2 * (1+xi) * (1-eta) * (1+eta)
        N7 = 1/2 * (1-xi) * (1+eta) * (1+xi)
        N8 = 1/2 * (1-xi) * (1-eta) * (1+eta)
        return N1, N2, N3, N4, N5, N6, N7, N8

    def coordinates(self, xi: float, eta: float) -> tuple[float, float]:
        """
        Calcula las coordenadas en un punto (xi, eta).
        """
        N1, N2, N3, N4, N5, N6, N7, N8 = self.shape_function(xi, eta)
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates
        x5, y5 = self.node5.vertex.coordinates
        x6, y6 = self.node6.vertex.coordinates
        x7, y7 = self.node7.vertex.coordinates
        x8, y8 = self.node8.vertex.coordinates
        x = N1*x1 + N2*x2 + N3*x3 + N4*x4 + N5*x5 + N6*x6 + N7*x7 + N8*x8
        y = N1*y1 + N2*y2 + N3*y3 + N4*y4 + N5*y5 + N6*y6 + N7*y7 + N8*y8
        return (x, y)

    def Ia(self) -> np.ndarray:
        """
        Calcula la matriz unitaria ampliada.
        """
        Ia = np.array([
            [1, 0, 0, 0],
            [0, 0, 0, 1],
            [0, 1, 1, 0]
        ])
        return Ia

    def J(self, xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz Jacobiana.
        """
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates
        x5, y5 = self.node5.vertex.coordinates
        x6, y6 = self.node6.vertex.coordinates
        x7, y7 = self.node7.vertex.coordinates
        x8, y8 = self.node8.vertex.coordinates

        dx_dxi = -0.25*eta**2*x1 + 0.25*eta**2*x2 + 0.25*eta**2*x3 - 0.25*eta**2*x4 - 0.5*eta**2*x6 + 0.5*eta**2*x8 - 0.5*eta*x1*xi + 0.25*eta*x1 - 0.5*eta*x2*xi - 0.25*eta*x2 + 0.5 * \
            eta*x3*xi + 0.25*eta*x3 + 0.5*eta*x4*xi - 0.25*eta*x4 + 1.0*eta*x5*xi - 1.0*eta*x7*xi + \
            0.5*x1*xi + 0.5*x2*xi + 0.5*x3*xi + 0.5*x4 * \
            xi - 1.0*x5*xi + 0.5*x6 - 1.0*x7*xi - 0.5*x8

        dy_dxi = -0.25*eta**2*y1 + 0.25*eta**2*y2 + 0.25*eta**2*y3 - 0.25*eta**2*y4 - 0.5*eta**2*y6 + 0.5*eta**2*y8 - 0.5*eta*xi*y1 - 0.5*eta*xi*y2 + 0.5*eta*xi*y3 + 0.5*eta*xi*y4 + \
            1.0*eta*xi*y5 - 1.0*eta*xi*y7 + 0.25*eta*y1 - 0.25*eta*y2 + 0.25*eta*y3 - 0.25*eta*y4 + \
            0.5*xi*y1 + 0.5*xi*y2 + 0.5*xi*y3 + 0.5*xi * \
            y4 - 1.0*xi*y5 - 1.0*xi*y7 + 0.5*y6 - 0.5*y8

        dx_deta = -0.5*eta*x1*xi + 0.5*eta*x1 + 0.5*eta*x2*xi + 0.5*eta*x2 + 0.5*eta*x3*xi + 0.5*eta*x3 - 0.5*eta*x4*xi + 0.5*eta*x4 - 1.0*eta*x6*xi - 1.0*eta*x6 + 1.0*eta*x8*xi - \
            1.0*eta*x8 - 0.25*x1*xi**2 + 0.25*x1*xi - 0.25*x2*xi**2 - 0.25*x2*xi + 0.25*x3*xi**2 + \
            0.25*x3*xi + 0.25*x4*xi**2 - 0.25*x4*xi + 0.5 * \
            x5*xi**2 - 0.5*x5 - 0.5*x7*xi**2 + 0.5*x7

        dy_deta = -0.5*eta*xi*y1 + 0.5*eta*xi*y2 + 0.5*eta*xi*y3 - 0.5*eta*xi*y4 - 1.0*eta*xi*y6 + 1.0*eta*xi*y8 + 0.5*eta*y1 + 0.5*eta*y2 + 0.5*eta*y3 + 0.5*eta*y4 - 1.0*eta*y6 - \
            1.0*eta*y8 - 0.25*xi**2*y1 - 0.25*xi**2*y2 + 0.25*xi**2*y3 + 0.25*xi**2*y4 + 0.5*xi**2 * \
            y5 - 0.5*xi**2*y7 + 0.25*xi*y1 - 0.25*xi*y2 + \
            0.25*xi*y3 - 0.25*xi*y4 - 0.5*y5 + 0.5*y7

        return np.array([[dx_dxi,  dy_dxi],
                        [dx_deta,  dy_deta]])

    def Gamma(self, xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz Jacobiana invertida.
        """
        J = self.J(xi, eta)
        gamma = np.linalg.inv(J)
        zero = np.zeros((2, 2))
        Gamma = np.concatenate((
            np.concatenate((gamma, zero), axis=1),
            np.concatenate((zero, gamma), axis=1)
        ),
            axis=0)
        return Gamma

    def dN(self, xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz de derivadas de las funciones de forma.
        """
        dN1_dxi = 0.25*(-eta - 2*xi)*(eta - 1)
        dN2_dxi = 0.25*(eta - 1)*(eta - 2*xi)
        dN3_dxi = 0.25*(eta + 1)*(eta + 2*xi)
        dN4_dxi = 0.25*(-eta + 2*xi)*(eta + 1)
        dN5_dxi = 1.0*xi*(eta - 1)
        dN6_dxi = 0.5 - 0.5*eta**2
        dN7_dxi = 1.0*xi*(-eta - 1)
        dN8_dxi = 0.5*eta**2 - 0.5
        dN1_deta = 0.25*(-2*eta - xi)*(xi - 1)
        dN2_deta = 0.25*(2*eta - xi)*(xi + 1)
        dN3_deta = 0.25*(2*eta + xi)*(xi + 1)
        dN4_deta = 0.25*(-2*eta + xi)*(xi - 1)
        dN5_deta = 0.5*xi**2 - 0.5
        dN6_deta = 1.0*eta*(-xi - 1)
        dN7_deta = 0.5 - 0.5*xi**2
        dN8_deta = 1.0*eta*(xi - 1)

        return np.array([
            [dN1_dxi,   0,        dN2_dxi,     0,           dN3_dxi,     0,           dN4_dxi,     0,
                dN5_dxi,     0,           dN6_dxi,     0,           dN7_dxi,     0,           dN8_dxi,     0],
            [dN1_deta,  0,        dN2_deta,    0,           dN3_deta,    0,           dN4_deta,    0,
                dN5_deta,    0,           dN6_deta,    0,           dN7_deta,    0,           dN8_deta,    0],
            [0,         dN1_dxi,  0,           dN2_dxi,     0,           dN3_dxi,     0,           dN4_dxi,
                0,           dN5_dxi,     0,           dN6_dxi,     0,           dN7_dxi,     0,           dN8_dxi],
            [0,         dN1_deta, 0,           dN2_deta,    0,           dN3_deta,    0,           dN4_deta,
                0,           dN5_deta,    0,           dN6_deta,    0,           dN7_deta,    0,           dN8_deta],
        ])

    def B(self, xi: float, eta: float) -> np.ndarray:
        """
        Calcula la matriz de deformacion.
        """
        dN = self.dN(xi, eta)
        Gamma = self.Gamma(xi, eta)
        Ia = self.Ia()
        B = Ia @ Gamma @ dN
        return B

    def D(self) -> np.ndarray:
        """
        Matriz constitutiva de esfuerzo plano.
        """
        v = self.section.v()
        E = self.section.E()
        # ! T E N S I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRESS:
            k = E/(1-v**2)
            return k*np.array([
                            [1, v, 0],
                            [v, 1, 0],
                            [0, 0, (1-v)/2],
                            ])

        # ! D E F O R M A C I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRAIN:
            k = (E*(1-v))/((1 + v)*(1 - 2*v))
            return k * np.array([
                [1,                 v/(1-v), 0],
                [v/(1-v),           1,       0],
                [0,                 0,       (1-2*v)/(2*(1-v))],
            ])

    def phiKi(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz funcional de rigidez.
        """
        detJ = np.linalg.det(self.J(xi, eta))
        B = self.B(xi, eta)
        D = self.D()
        t = self.t
        return B.T @ D @ B * detJ * t

    @staticmethod
    def P() -> np.ndarray:
        """
        Matriz de Permutacion.
        """
        P = np.array([
            [0, 1, 0, 0, 0, 0, 0, 0],
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 1, 0]
            ])
        return P

    def global_stiffness_matrix(self) -> np.ndarray:
        """
        Calcula la matriz rigidez global atravez de la intergracion numerica cuadratura de gauss.
        y condensa la matriz rigidez en los nodos de vertices.
        """
        Ki = np.zeros((16, 16))
        for i in range(len(self.xi)):
            Ki += self.phiKi(self.xi[i], self.eta[i]) * self.w[i]
        # Condensacion estatica:
        # kc =kcc-kci*kii^-1*kic
        c = [0, 1, 2, 3, 4, 5, 6, 7]
        i = [8, 9, 10, 11, 12, 13, 14, 15]
        Kcc = Ki[np.ix_(c, c)]
        Kci = Ki[np.ix_(c, i)]
        Kic = Ki[np.ix_(i, c)]
        Kii = Ki[np.ix_(i, i)]
        kc = Kcc - Kci @ np.linalg.inv(Kii) @ Kic
        # kc = self.P() @ kc @ self.P().T
        return kc
        # return Ki


if __name__ == "__main__":
    import pandas as pd
    from milcapy.utils.geometry import Vertex
    pd.set_option('display.float_format', '{:.7f}'.format)
    from milcapy.material.material import Material
    from milcapy.section.section import ShellSection
    # Units: tonf, m
    l = 1.5  # m
    h = 0.6  # m
    E = 2534.56*10**3  # tonf/m2
    v = 0.2
    t = 0.25  # m
    material = Material("material", E, v, 0)
    section = ShellSection("section", material, t)
    nd1 = Node(1, Vertex(0, 0))
    nd2 = Node(2, Vertex(h, 0))
    nd3 = Node(3, Vertex(h, l))
    nd4 = Node(4, Vertex(0, l))
    ele = MembraneQuad8(1, nd1, nd2, nd3, nd4, section, ConstitutiveModel.PLANE_STRESS, IntegrationType.REDUCED)
    # K = ele.Ki()/10**5#[4:16, 4:16] #*0.018
    # F = np.array([3, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    # u = np.linalg.solve(K, F)
    K = ele.global_stiffness_matrix()[4:8, 4:8]
    F = np.array([3, 0, 3, 0])
    u = np.linalg.solve(K, F)
    # print(pd.DataFrame(K))
    print(pd.DataFrame(u)*1000)
    # print(ele.D())



import numpy as np
from typing import TYPE_CHECKING, Optional, Dict, Tuple
from milcapy.utils.geometry import angle_x_axis
from milcapy.loads.load import DistributedLoad
from milcapy.utils.types import BeamTheoriesType
from milcapy.utils.element import (
    local_stiffness_matrix,
    transformation_matrix,
    length_offset_transformation_matrix,
    length_offset_q,
    q_phi
)

from milcapy.loads.load import FrameRelease

if TYPE_CHECKING:
    from milcapy.core.node import Node
    from milcapy.utils.types import MemberType
    from milcapy.section.section import Section


class Member:
    """
    Clase que representa un miembro estructural.
    """

    def __init__(
        self,
        id:           int,
        node_i:      "Node",
        node_j:      "Node",
        section:     "Section",
        member_type: "MemberType",
        beam_theory: "BeamTheoriesType",
    ) -> None:
        """
        Inicializa un objeto que representa un marco estructural.

        Args:
            id (int): Identificador del miembro.
            node_i (Node): Nodo inicial.
            node_j (Node): Nodo final.
            section (Section): Sección del miembro.
            member_type (MemberType): Tipo de miembro.
            beam_theory (BeamTheoriesType): Teoría de vigas.
        """

        self.id = id
        self.node_i = node_i
        self.node_j = node_j
        self.section = section
        self.member_type = member_type
        self.beam_theory = beam_theory

        # Mapa de grados de libertad
        self.dofs: np.ndarray = np.concatenate([node_i.dofs, node_j.dofs]) # [dofs node_i, dofs node_j]

        # Cargas distribuidas
        self.distributed_load: Dict[str, DistributedLoad] = {}  # {pattern_name: DistributedLoad}

        # Patrón de carga actual
        self.current_load_pattern: Optional[str] = None

        # Topicos
        self.la:  Optional[float] = None    # Longitud del brazo rigido inicial
        self.lb:  Optional[float] = None    # Longitud del brazo rigido final
        self.qla: Optional[bool]  = None    # Indica si hay cargas en el brazo rigido inicial
        self.qlb: Optional[bool]  = None    # Indica si hay cargas en el brazo rigido final
        self.fla: Optional[float] = None    # Factor de longitud del brazo rigido inicial
        self.flb: Optional[float] = None    # Factor de longitud del brazo rigido final

        # Liberaciones
        self.release: Optional[FrameRelease] = None

    def length(self) -> float:
        """
        Longitud del miembro.
        """
        return (self.node_i.vertex.distance_to(self.node_j.vertex))

    def angle_x(self) -> float:
        """
        Ángulo del miembro respecto al eje X del sistema global.
        """
        return angle_x_axis(
            self.node_j.vertex.x - self.node_i.vertex.x,
            self.node_j.vertex.y - self.node_i.vertex.y
        )

    def phi(self) -> float:
        """
        Ángulo de corte para efectos de deformación por cortante (parámetro phi).

        phi = (12 * E * I) / (L**2 * A * k * G)
        """
        if self.beam_theory == BeamTheoriesType.EULER_BERNOULLI:
            return 0

        elif self.beam_theory == BeamTheoriesType.TIMOSHENKO:

            laef, lbef = self.la_lb_efect()
            E = self.section.E()
            I = self.section.I()
            L = self.length() - laef - lbef
            A = self.section.A()
            k = self.section.k()
            G = self.section.G()

            return (12 * E * I) / (L**2 * A * k * G) if k != 0 else 0

    def la_lb_efect(self) -> Tuple[float, float]:
        """
        Longitudes efectivas de los brazos rigidos.
        """
        la = self.la or 0
        fla = self.fla or 1
        laef = la * fla

        lb = self.lb or 0
        flb = self.flb or 1
        lbef = lb * flb

        return laef, lbef

    def set_current_load_pattern(self, load_pattern_name: str) -> None:
        """
        Establece el patrón de carga actual del miembro.
        """
        self.current_load_pattern = load_pattern_name

    def set_distributed_load(self, load: DistributedLoad) -> None:
        """
        Asigna una carga distribuida al miembro para el patrón de carga actual.

        Args:
            load (DistributedLoad): Carga distribuida a aplicar.
        """
        if self.current_load_pattern is None:
            raise ValueError("Debe establecer un patrón de carga actual antes de asignar una carga distribuida.")
        self.distributed_load[self.current_load_pattern] = load

    def get_distributed_load(self, load_pattern_name: str) -> Optional["DistributedLoad"]:
        """
        Obtiene la carga distribuida para el patrón de carga actual.

        Args:
            load_pattern_name (str): Nombre del patrón de carga.

        Returns:
            Optional["DistributedLoad"]: Carga distribuida.
        """
        load = self.distributed_load.get(load_pattern_name, None)
        if load is None:
            return DistributedLoad()
        return load

    def transformation_matrix(self) -> np.ndarray:
        """
        Calcula la matriz de transformación del miembro.

        Returns:
            np.ndarray: Matriz de transformación.
        """
        return transformation_matrix(
            angle=self.angle_x()
        )

    def length_offset_transformation_matrix(self) -> np.ndarray:
        """
        Matriz de transformación para miembros con brazos rígidos (identidad si no hay brazos).
        """
        if self.la is None and self.lb is None:
            return np.eye(6)  # Matriz identidad 6x6
        laef, lbef = self.la_lb_efect()
        H = length_offset_transformation_matrix(laef, lbef)
        return H

    def k_local(self) -> np.ndarray:
        """
        Calcula la matriz de rigidez local del miembro.

        Returns:
            np.ndarray: Matriz de rigidez en local.
        """
        if self.la is None and self.lb is None:
            k = local_stiffness_matrix(
                E=self.section.E(),
                I=self.section.I(),
                A=self.section.A(),
                L=self.length(),
                le=self.length(),
                phi=self.phi(),
            )
            return k
        else:
            laef, lbef = self.la_lb_efect()
            H = self.length_offset_transformation_matrix()
            k = local_stiffness_matrix(
                E=self.section.E(),
                I=self.section.I(),
                A=self.section.A(),
                L=self.length(),
                le=self.length() - laef - lbef,
                phi=self.phi(),
            )
            return H.T @ k @ H

    def local_stiffness_matrix(self) -> np.ndarray:
        """
        Calcula la matriz de rigidez local del miembro.

        Returns:
            np.ndarray: Matriz de rigidez en local.
        """
        k = self.k_local()
        if self.release is not None:
            freeDof = self.release.get_dof_release()
            for dof in freeDof:
                k = self.apply_releases_for_stiffness_matrix(dof=dof, k=k)
        return k

    def apply_releases_for_stiffness_matrix(self, dof: int, k: np.ndarray) -> np.ndarray:
        """
        Calcula la matriz de rigidez local del miembro con liberaciones.

        Args:
            dof (int): Grado de libertad a liberar.
            k (np.ndarray): Matriz de rigidez local del miembro.

        Returns:
            np.ndarray: Matriz de rigidez local del miembro con liberaciones.
        """

        # Todos los DOF posibles en 6x6
        allDof = np.arange(6)

        # Submatrices
        k1 = k[np.ix_(allDof, [dof])]
        k2 = k[np.ix_([dof], allDof)]
        k3 = k[np.ix_([dof], [dof])]

        # Reducción de rigidez con liberaciones
        kp = k - k1 @ (1/k3) @ k2
        return kp


    def flexible_stiffness_matrix(self) -> np.ndarray:
        """
        Calcula la matriz de rigidez de la parte flexible del miembro.

        Returns:
            np.ndarray: Matriz de rigidez de la parte flexible del miembro.
        """
        laef, lbef = self.la_lb_efect()
        k = local_stiffness_matrix(
            E=self.section.E(),
            I=self.section.I(),
            A=self.section.A(),
            L=self.length(),
            le=self.length() - laef - lbef,
            phi=self.phi(),
        )
        return k

    def q_local(self) -> np.ndarray:
        """
        Calcula el vector de cargas distribuidas equivalentes en sistema local.

        Returns:
            np.ndarray: Vector de cargas distribuidas equivalentes en sistema local.
        """
        load = self.get_distributed_load(self.current_load_pattern)
        if self.la is None and self.lb is None:
            q = q_phi(
                L=self.length(),
                phi=self.phi(),
                qi=load.q_i,
                qj=load.q_j,
                pi=load.p_i,
                pj=load.p_j
            )
            return q
        else:
            laef, lbef = self.la_lb_efect()
            q = length_offset_q(
                self.length(),
                self.phi(),
                load.q_i,
                load.q_j,
                load.p_i,
                load.p_j,
                laef,
                lbef,
                self.qla,
                self.qlb,
            )
            return q

    def local_load_vector(self) -> np.ndarray:
        """
        Calcula el vector de cargas distribuidas equivalentes en sistema local.

        Returns:
            np.ndarray: Vector de cargas distribuidas equivalentes en sistema local.
        """
        q = self.q_local()
        k = self.k_local()
        if self.release is not None:
            freeDof = self.release.get_dof_release()
            for dof in freeDof:
                q = self.apply_releases_for_load_vector(dof=dof, q=q, k=k)
                k = self.apply_releases_for_stiffness_matrix(dof=dof, k=k)
        return q

    def apply_releases_for_load_vector(
        self,
        dof: int,
        q: np.ndarray,
        k: np.ndarray
    ) -> np.ndarray:
        """
        Aplica liberaciones al vector de cargas.

        Args:
            dof (int): Grado de libertad a liberar.
            q (np.ndarray): Vector de cargas.
            k (np.ndarray): Matriz de rigidez local del miembro.

        Returns:
            np.ndarray: Vector de cargas con liberaciones.
        """
        vecudof = np.zeros_like(q)
        if k[dof, dof] != 0.0:
            vecudof[dof] = q[dof] / k[dof, dof]
        else:
            raise ZeroDivisionError(f"El término k[{dof},{dof}] es cero, no se puede aplicar release.")

        # Nuevo vector de cargas con liberación
        qp = q - k @ vecudof
        return qp



    def global_stiffness_matrix(self) -> np.ndarray:
        """
        Calcula la matriz de rigidez global del miembro.

        Returns:
            np.ndarray: Matriz de rigidez global.
        """
        T = self.transformation_matrix()
        ke = self.local_stiffness_matrix()

        return T.T @ ke @ T

    def global_load_vector(self) -> np.ndarray:
        """
        Calcula el vector de fuerzas global del miembro para el patron de carga actual.

        Returns:
            np.ndarray: Vector de fuerzas global.
        """
        if self.current_load_pattern is None:
            raise ValueError("Debe establecer un patrón de carga actual antes de calcular el vector de fuerzas global.")
        T = self.transformation_matrix()
        q = self.local_load_vector()

        return T.T @ q

    def add_releases(
        self,
        uxi: bool=False,
        uyi: bool=False,
        rzi: bool=False,
        uxj: bool=False,
        uyj: bool=False,
        rzj: bool=False
    ) -> None:
        """
        Asigna liberaciones a un miembro.

        Args:
            uxi (bool, opcional): Liberación de U en el nodo inicial en la dirección x. Default es False.
            uyi (bool, opcional): Liberación de U en el nodo inicial en la dirección y. Default es False.
            rzi (bool, opcional): Liberación de R en el nodo inicial en la dirección z. Default es False.
            uxj (bool, opcional): Liberación de U en el nodo final en la dirección x. Default es False.
            uyj (bool, opcional): Liberación de U en el nodo final en la dirección y. Default es False.
            rzj (bool, opcional): Liberación de R en el nodo final en la dirección z. Default es False.
        """
        self.release = FrameRelease(uxi=uxi, uyi=uyi, rzi=rzi, uxj=uxj, uyj=uyj, rzj=rzj)


from .model.model import SystemMilcaModel
from .utils.types import (
    BeamTheoriesType,
    CoordinateSystemType,
    DirectionType,
    StateType,
    LoadType,
    FieldTypeMembrane,
    ConstitutiveModel,
    IntegrationType,
    )

class SystemModel(SystemMilcaModel):
    """
    Class to create a model for the MILCA software.

    Attributes:
        model (SystemMilcaModel): Instance of the SystemMilcaModel class.
    """

    def __init__(self):
        super().__init__()

def model_viewer(
    model: SystemMilcaModel
):
    """
    Muestra la interfaz gráfica para visualizar el modelo.
    """
    model.show()